import os
import clisync
import shutil

from typing import Optional, List, Union

from pyhectiqlab.tag import Tag
from pyhectiqlab.project import Project
from pyhectiqlab.client import Client
from pyhectiqlab.decorators import functional_alias
from pyhectiqlab.utils import batched, list_all_files_in_dir, is_running_event_loop
from pyhectiqlab.logging import hectiqlab_logger as logger
from pyhectiqlab.settings import getenv


class Model:
    _client: Client = Client

    @staticmethod
    @functional_alias("create_model")
    @clisync.include()
    def create(
        name: str,
        path: Optional[str] = None,
        version: Optional[str] = None,
        run_id: Optional[str] = None,
        project: Optional[str] = None,
        description: Optional[str] = None,
        upload: bool = True,
    ):
        """Create a model

        Args:
            name (str): Name of the model
            path (str): Path to the model directory. If None, the model is created without files. Default: None.
            run_id (str, optional):  ID of the run
            project (str, optional): Project of the model
            version (str, optional): Version of the model
            description (str, optional): Description of the model
            upload (bool, optional): Upload the model files. Default: True

        Raises:
            logging.error: If the `name` parameter is not provided, the project is not found or the model creation fails, it does nothing and returns None.
        """
        from pyhectiqlab.run import Run

        if name is None:
            logger.error("The `name` parameter is required.")
            return

        project = Project.get(project)
        run_id = Run.get_id(run_id)

        if project is None:
            msg = f"Could not create the model {name}, project not found."
            if not Model._client.online():
                msg = msg[:-1] + " (offline mode)."
            logger.error(msg)
            return

        json = {
            "name": name,
            "description": description,
            "version": version,
            "project": project,
            "root_run": run_id,
        }

        model = Model._client.post("/app/models", wait_response=True, json=json)
        if model is None or "id" not in model:
            logger.error(f"Failed to create model {name}.")
            return

        if upload and path:
            Model.upload(id=model["id"], path=path)

        if run_id is not None:
            Model.attach(name=name, version=version, run_id=run_id, project=project, wait_response=True)
        return model

    @staticmethod
    @functional_alias("retrieve_model")
    @clisync.include()
    def retrieve(
        name: str,
        version: str,
        project: Optional[str] = None,
        fields: Optional[List[str]] = [],
    ):
        """Retrieve a model

        Args:
            name (str): Name of the model
            version (str): Version of the model
            project (str, optional): Project of the model. If None, the project of the current run is used. Default: None.
            fields (dict, optional): Fields to retrieve
            wait_response (bool, optional): Wait for the response from the server

        Raises:
            logging.error: If the project is not found, it does nothing and returns None.
        """
        project = Project.get(project)
        if project is None:
            msg = "Could not retrieve the model, project not found."
            if not Model._client.online():
                msg = msg[:-1] + " (offline mode)."
            msg += " Skipping model retrieval."
            logger.error(msg)
            return

        return Model._client.get(
            "/app/models/retrieve",
            params={
                "name": name,
                "project": project,
                "version": version,
                "fields": fields,
            },
            wait_response=True,
        )

    @staticmethod
    @functional_alias("model_exists")
    @clisync.include()
    def exists(name: str, version: str, project: Optional[str] = None):
        """Check if a model exists

        Args:
            name (str): Name of the model.
            version (str): Version of the model.
            project (str): Project of the model.
        """
        return Model.retrieve(name=name, project=project, version=version, fields=["id"]) is not None

    @staticmethod
    @functional_alias("list_models")
    @clisync.include(wait_response=True)
    def list(
        project: Optional[str] = None,
        search: Optional[str] = None,
        author: Optional[str] = None,
        keep_latest_version: bool = False,
        fields: Optional[List[str]] = [],
        page: Optional[int] = 1,
        limit: Optional[int] = 100,
        order_by: Optional[str] = "created_at",
        order_direction: Optional[str] = "desc",
        wait_response: bool = False,
    ):
        """List the models

        Args:
            project (str): Project of the model
            search (str, optional): Search string
            author (str, optional): Author of the model
            keep_latest_version (bool, optional): If True, group by the latest version of model name and return only the latest version of each model name
            fields (List[str], optional): Fields to retrieve
            page (int, optional): Page number
            limit (int, optional): Limit of the models
            order_by (str, optional): Order by
            order_direction (str, optional): Order direction
            wait_response (bool, optional): Wait for the response from the server
        """

        params = {
            "project": Project.get(project),
            "search": search,
            "author": author,
            "keep_latest_version": keep_latest_version,
            "fields": fields,
            "page": page or 1,
            "limit": limit or 100,
            "order_by": order_by,
            "order_direction": order_direction,
        }
        return Model._client.get("/app/models", params=params, wait_response=wait_response)

    @staticmethod
    @functional_alias("update_model")
    @clisync.include(wait_response=True)
    def update(
        id: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        version: Optional[str] = None,
        wait_response: bool = False,
    ):
        """Update a model

        Args:
            id (str): ID of the model
            name (str, optional): Name of the model
            description (str, optional): Description of the model
            version (str, optional): Version of the model
            wait_response (bool, optional): Wait for the response from the server

        Raises:
            If the `id` parameter is not provided, it does nothing and returns None.
        """
        if not id:
            logger.error("The `id` parameter is required.")
            return

        return Model._client.put(
            f"/app/models/{id}",
            json={"name": name, "description": description, "version": version},
            wait_response=wait_response,
        )

    @staticmethod
    @functional_alias("delete_model")
    @clisync.include(wait_response=True)
    def delete(
        id: Optional[str] = None,
        name: Optional[str] = None,
        version: Optional[str] = None,
        project: Optional[str] = None,
        wait_response: bool = False,
    ):
        """Delete a model

        Args:
            id (str): ID of the model
            name (str, optional): Name of the model
            version (str, optional): Version of the model
            project (str, optional): Project of the model
            wait_response (bool, optional): Wait for the response from the server

        Raises:
            logger.error: If the `id` parameter is not provided, or all of `name`, `version`, and `project` are not provided, it does nothing and returns None.
        """
        project = Project.get(project)
        if id is None and (name is None or version is None or project is None):
            raise ValueError("Either `id` or `name`, `version` and `project` are required.")
        if id is None:
            model = Model.retrieve(name=name, project=project, version=version, fields=["id"])
            if model is None:
                logger.error(f"Could not delete model {name}-{version}, model not found. Skipping model deletion.")
                return
            id = model["id"]
        return Model._client.delete(f"/app/models/{id}", wait_response=wait_response)

    @staticmethod
    @clisync.include()
    @functional_alias("upload_model")
    @clisync.include()
    def upload(id: str, path: str) -> None:
        """Upload local files to a model hosted in the Lab.

        Args:
            id (str): ID of the model.
            path (str): Path to the file or directory containing the files to upload.
        """
        all_files = list_all_files_in_dir(path) if os.path.isdir(path) else [path]
        batch_size = 50
        Model._id = model_id = id

        def composition_sync(**kwargs):
            for batch in batched(all_files, batch_size):
                # Create many files
                batch_body = [{"name": os.path.relpath(f, start=path if os.path.isdir(path) else os.path.dirname(f)), 
                               "num_bytes": os.path.getsize(f)} for f in batch]
                files = Model._client.post(
                    f"/app/models/{model_id}/files", wait_response=True, json={"files": batch_body}
                )["results"]

                # For each file in batch, get the policy in files (use name for finding the file)
                sorted_files = []
                for el in batch:
                    name = os.path.relpath(el, start=path if os.path.isdir(path) else os.path.dirname(el))
                    file = [
                        f.get("upload_policy")
                        for f in files
                        if f["name"]
                        == name
                    ][0]
                    sorted_files.append(file)
                # Upload dataset files
                Model._client.upload_many_sync(paths=batch, policies=sorted_files, **kwargs)

        async def composition_async(**kwargs):
            for batch in batched(all_files, batch_size):
                # Create many files
                batch_body = [{"name": os.path.relpath(f, start=path if os.path.isdir(path) else os.path.dirname(f)), 
                               "num_bytes": os.path.getsize(f)} for f in batch]
                files = Model._client.post(
                    f"/app/models/{model_id}/files", wait_response=True, json={"files": batch_body}
                )["results"]

                # For each file in batch, get the policy in files (use name for finding the file)
                sorted_files = []
                for el in batch:
                    name = os.path.relpath(el, start=path if os.path.isdir(path) else os.path.dirname(el))
                    file = [f.get("upload_policy") for f in files if f["name"] == name][0]
                    sorted_files.append(file)
                # Upload dataset files
                await Model._client.upload_many_async(paths=batch, policies=sorted_files, **kwargs)

        if is_running_event_loop():
            Model._client.execute(composition_sync, wait_response=True, is_async_method=False, with_progress=True)
        else:
            Model._client.execute(composition_async, wait_response=True, is_async_method=True, with_progress=True)

    @staticmethod
    @functional_alias("download_model")
    @clisync.include()
    def download(
        name: str,
        version: str,
        project: Optional[str] = None,
        path: Optional[str] = None,
        overwrite: bool = False,
    ):
        """Download a model

        Args:
            name (str): Name of the model
            project (str): Project of the model
            version (str): Version of the model
            path (str, optional): Path to download the model
            overwrite (bool, optional): Overwrite the files if they already exist

        Raises:
            logging.error: If the project is not found or the model does not exist, it does nothing and returns None.
        """
        from pyhectiqlab import Run
        project = Project.get(project)
        if project is None:
            msg = "Could not download the model, project not found."
            if not Model._client.online():
                msg = msg[:-1] + "(offline mode)."
            logger.error(msg)
            return
        if not Model.exists(name=name, version=version, project=project):
            logger.error(
                f"Could not download the model, model {name}-{version} does not exist. Skipping model download."
            )
            return
        
        run_id = Run.get_id()
        if run_id is not None:
            Model.attach(name=name, version=version, run_id=run_id, project=project)


        path = path or getenv("HECTIQLAB_MODELS_DOWNLOAD") or "./"
        if not os.path.basename(path) == f"{name}-{version}":
            path = os.path.join(path, f"{name}-{version}")
        if os.path.exists(path):
            if not overwrite:
                logger.warning(
                    f"Directory {path} already exists. Set `overwrite=True` to overwrite the files. Skipping model download."
                )
                return path

            shutil.rmtree(path)
        os.makedirs(path, exist_ok=True)

        project = Project.get(project)

        def composition_sync(**kwargs):
            model = Model.retrieve(name=name, project=project, version=version, fields=["id"])
            if not model:
                return
            model_id = model["id"]

            num_results = 0
            total_results = 1
            page = 1
            while not num_results == total_results:
                files = Model._client.get(
                    f"/app/models/{model_id}/files/",
                    params={"page": page, "limit": 50, "fields": ["download_url", "name", "num_bytes"]},
                    wait_response=True,
                )
                total_results = files["total_results"]

                Model._client.download_many_sync(
                    urls=[file["download_url"] for file in files["results"]],
                    local_paths=[os.path.join(path, file["name"]) for file in files["results"]],
                    num_bytes=[file["num_bytes"] for file in files["results"]],
                    **kwargs,
                )
                num_results += len(files["results"])
                page += 1

            Model._client.post(f"/app/models/{model_id}/on-download-completed", wait_response=False)

        async def composition_async(**kwargs):
            model = Model.retrieve(name=name, project=project, version=version, fields=["id"])
            if not model:
                return
            model_id = model["id"]

            num_results = 0
            total_results = 1
            page = 1
            while not num_results == total_results:
                files = Model._client.get(
                    f"/app/models/{model_id}/files/",
                    params={"page": page, "limit": 50, "fields": ["download_url", "name", "num_bytes"]},
                    wait_response=True,
                )
                total_results = files["total_results"]

                await Model._client.download_many_async(
                    urls=[file["download_url"] for file in files["results"]],
                    local_paths=[os.path.join(path, file["name"]) for file in files["results"]],
                    num_bytes=[file["num_bytes"] for file in files["results"]],
                    **kwargs,
                )
                num_results += len(files["results"])
                page += 1

            Model._client.post(f"/app/models/{model_id}/on-download-completed", wait_response=False)

        logger.info(f"⏳ Downloading model {name}-{version} to {path}...")
        if is_running_event_loop():
            Model._client.execute(composition_sync, wait_response=True, is_async_method=False, with_progress=True)
        else:
            Model._client.execute(composition_async, wait_response=True, is_async_method=True, with_progress=True)

        logger.info("✅ Done.")
        return path

    @staticmethod
    def attach(
        name: str,
        version: str,
        run_id: str,
        project: Optional[str] = None,
        wait_response: bool = False,
    ):
        """Attach a model to a run

        Args:
            run_id (str): ID of the run.
            name (str): Name of the model.
            version (str): Version of the model.
            project (str, optional): Project of the model. If None, the project of the current run is used. Default: None.
            wait_response (bool, optional): Wait for the response from the server. Default: False.

        Raises:
            logging.error: If the `run_id` parameter is not provided, or if the model does not exist, it does nothing and returns None.

        """

        if run_id is None:
            logger.error("The `run_id` parameter is required.")
            return
        project = Project.get(project)
        model = Model.retrieve(name=name, project=project, version=version, fields=["id"])
        if not model:
            logger.error(
                f"Could not attach run {run_id}, model {name}-{version} not found. Skipping attachment to run."
            )
            return
        model_id = model["id"]
        return Model._client.post(f"/app/models/{model_id}/attach-to-run/{run_id}", wait_response=wait_response)

    @staticmethod
    def detach(
        name: str,
        version: str,
        run_id: str,
        project: Optional[str] = None,
        wait_response: bool = False,
    ):
        """Detach a model from a run

        Args:
            name (str): Name of the model.
            version (str): Version of the model.
            run_id (int): ID of the run.
            project (str, optional): Project of the model. If None, the project of the current run is used. Default: None.
            wait_response (bool, optional): Wait for the response from the server. Default: False.

        Raises:
            logging.error: If the model is not found, it does nothing and returns None.
        """
        project = Project.get(project)
        model = Model.retrieve(name=name, project=project, version=version, fields=["id"])
        if not model:
            logger.error(
                f"Could not detach run {run_id}, model {name}-{version} not found. Skipping detachment from run."
            )
            return
        model_id = model["id"]
        return Model._client.post(f"/app/models/{model_id}/detach-from-run/{run_id}", wait_response=wait_response)

    @staticmethod
    @functional_alias("add_tags_to_model")
    @clisync.include()
    def add_tags(
        tags: Union[str, List[str]],
        name: str,
        version: str,
        project: Optional[str] = None,
    ):
        """Add tags to a model

        Args:
            tags (Union[str, List[str]]): Tags to add.
            name (str): Name of the model.
            version (str): Version of the model.
            project (str): Project of the model. If None, the project of the current run is used. Default: None.

        Raises:
            logging.error: If the model is not found, it does nothing and returns None.
        """
        project = Project.get(project)
        model = Model.retrieve(name=name, project=project, version=version, fields=["id"])
        if not model:
            logger.error(
                f"Could not attach tags {tags}, model {name}-{version} not found. Skipping attachment of tags."
            )
            return
        model_id = model["id"]
        return Tag.attach_to_model(tags=tags, model_id=model_id, project=project)

    @staticmethod
    @functional_alias("detach_tag_from_model")
    @clisync.include()
    def detach_tag(
        tag: str,
        name: str,
        version: str,
        project: Optional[str] = None,
    ):
        """Remove a tag from the model.

        Args:
            title (str): The new title of the run.
            name (str): Name of the model.
            version (str): Version of the model.
            project (str): Project of the model. If None, the project of the current run is used. Default: None.

        Raises:
            logging.error: If the model is not found, it does nothing and returns
        """
        project = Project.get(project)
        model = Model.retrieve(name=name, version=version, project=project, fields=["id"])
        if not model:
            logger.error(f"Could not detach tag {tag}, model {name}-{version} not found. Skipping detachment of tag.")
            return
        model_id = model["id"]
        Tag.detach_from_model(tag=tag, model_id=model_id, project=project)



import os
import logging
from typing import Any, Optional
from dotenv import load_dotenv


def load_env(override: bool = False):
    """Load environment variables from .env files. This function loads the public
    environment variables from a `.env.public` file and then loads specific
    environment variables from a `.env.{ENV}` file where `{ENV}` is the value of the
    `ENV` environment variable.

    Args:
        override (bool): Whether to override existing environment variables.
            Default: False.
    """
    # Load comments from .env file
    env_path = os.path.join(os.path.dirname(__file__), ".env.public")
    if not env_path:
        assert False, f"🚫 api/app/settings: Could not find .env.public file"
    else:
        logging.info(f"✅ Loading secrets at {env_path}")
        load_dotenv(env_path, override=override)

    # Load specific environment variables from .env.{ENV} file
    ENV = os.getenv("ENV", "local")
    env_path = os.path.join(os.path.dirname(__file__), f".env.{ENV}")
    if not env_path:
        logging.info(f"🚫 api/app/settings:::.env.{ENV} file")
    else:
        logging.info(f"✅ Loading secrets at {env_path}")
        load_dotenv(env_path, override=override)


def getenv(key: str, default: Optional[Any] = None) -> Any:
    """Get environment variable or return explicit default. This function overrides
    the behavior of os.getenv() to return a boolean or None if the value is set to
    "True", "False", or "None" respectively.

    Args:
        key (str): The environment variable key.
        default (Any, optional): The default value to return if the key is not found.
            Defaults to None.

    Returns:
        Any: The environment variable value or the default value.
    """
    value = os.getenv(key, default)
    if value == "True":
        return True
    elif value == "False":
        return False
    elif value == "None":
        return None
    return value



import httpx
import asyncio
import threading
import logging
import os
import io
from contextlib import contextmanager

from typing import Any, Optional, List, Type, Union
from functools import partial
from google.resumable_media.requests import ResumableUpload


from pyhectiqlab.auth import Auth
from pyhectiqlab.decorators import classproperty, request_handle, execute_online_only
from pyhectiqlab.settings import getenv
from pyhectiqlab.utils import ProgressHandler, is_running_event_loop

from pyhectiqlab import API_URL


logger = logging.getLogger()

ResponseType = Union[dict[str, Any], bytes, Type[None]]


class Client:
    """
    Client singleton for making sync and async requests in the Hectiq Lab API. This client
    can be used in a context manager, or as a singleton. For performing async requests, the object
    must be instanciated.

    Example:
    ```python
    from pyhectiqlab.client import Client
    import asyncio

    client = Client()

    # Sync request
    response = client.get("/app/auth/is-logged-in")

    # Async request
    response = asyncio.run(client.async_get("/app/auth/is-logged-in"))
    ```
    """

    _auth: Auth = None
    _client: httpx.Client = None
    _online: bool = True
    _timeout: float = 30.0  # seconds
    _progress: Optional[ProgressHandler] = None
    _logged_error_messages: set[str] = set()

    @classproperty
    def auth(cls) -> Auth:
        if cls._auth is None:
            cls._auth = Auth()
        return cls._auth

    @classproperty
    def progress(cls) -> ProgressHandler:
        if cls._progress is None:
            cls.initiate_progress()
        return cls._progress

    @classproperty
    def client(cls) -> httpx.Client:
        if cls._client is None:
            cls._client = httpx.Client(auth=cls.auth, timeout=cls._timeout)
        return cls._client

    @staticmethod
    def is_logged():
        return Client.auth.is_logged()

    @staticmethod
    def online(status: Optional[bool] = None):
        if status is not None and not is_running_event_loop():
            Client._online = status
        if getenv("HECTIQLAB_OFFLINE_MODE"):
            return False
        return Client.auth.online and Client._online

    @staticmethod
    def get(url: str, wait_response: bool = False, **kwargs: Any) -> ResponseType:
        return Client.request("get", url, wait_response=wait_response, **kwargs)

    @staticmethod
    @execute_online_only
    def post(url: str, wait_response: bool = False, **kwargs: Any) -> ResponseType:
        return Client.request("post", url, wait_response=wait_response, **kwargs)

    @staticmethod
    @execute_online_only
    def patch(url: str, wait_response: bool = False, **kwargs: Any) -> ResponseType:
        return Client.request("patch", url, wait_response=wait_response, **kwargs)

    @staticmethod
    @execute_online_only
    def put(url: str, wait_response: bool = False, **kwargs: Any) -> ResponseType:
        return Client.request("put", url, wait_response=wait_response, **kwargs)

    @staticmethod
    @execute_online_only
    def delete(url: str, wait_response: bool = False, **kwargs: Any) -> ResponseType:
        return Client.request("delete", url, wait_response=wait_response, **kwargs)

    @staticmethod
    @execute_online_only
    def upload_sync(local_path: str, policy: dict) -> ResponseType:
        """Upload a file."""
        upload_method = policy.get("upload_method")
        if not upload_method:
            return

        if upload_method == "fragment":
            res = Client.upload_fragment_sync(local_path, policy)
        elif upload_method == "single":
            res = Client.upload_single_sync(local_path, policy.get("policy"), bucket=policy.get("bucket_name"))
        return res

    @staticmethod
    @execute_online_only
    async def upload_async(local_path: str, policy: dict) -> ResponseType:
        """Upload a file."""
        upload_method = policy.get("upload_method")
        if not upload_method:
            return

        async with httpx.AsyncClient(timeout=Client._timeout) as asyncClient:
            if upload_method == "fragment":
                res = await Client.upload_fragment_async(local_path, policy, client=asyncClient)
            elif upload_method == "single":
                res = await Client.upload_single_async(
                    local_path,
                    policy.get("policy"),
                    bucket=policy.get("bucket_name"),
                    client=asyncClient,
                )
        return res

    @staticmethod
    def get_upload_method(policy: dict) -> dict[str, Any]:
        """Get the upload method from the policy."""
        if "upload_method" in policy:
            return policy.get("upload_method")
        return policy.get("policy", {}).get("upload_method")

    @staticmethod
    @execute_online_only
    def upload_many_sync(paths: List[str], policies: List[dict]) -> None:
        """Upload many files synchronously."""
        single_upload_files = []
        fragment_upload_files = []
        for path, policy in zip(paths, policies):
            if not policy:
                continue
            method = Client.get_upload_method(policy)
            if method == "fragment":
                fragment_upload_files.append((path, policy))
            elif method == "single":
                single_upload_files.append((path, policy))

        for path, policy in single_upload_files:
            Client.upload_single_sync(path, policy.get("policy"), bucket=policy.get("bucket_name"))
        for path, policy in fragment_upload_files:
            Client.upload_fragment_sync(path, policy)

    @staticmethod
    @execute_online_only
    async def upload_many_async(paths: List[str], policies: List[dict]) -> None:
        """Upload many files asynchronously."""
        single_upload_files = []
        fragment_upload_files = []
        for path, policy in zip(paths, policies):
            if not policy:
                continue
            method = Client.get_upload_method(policy)
            if method == "fragment":
                fragment_upload_files.append((path, policy))
            elif method == "single":
                single_upload_files.append((path, policy))

        async with httpx.AsyncClient() as client:
            tasks = []
            for path, policy in single_upload_files:
                task = Client.upload_single_async(
                    path, policy.get("policy"), bucket=policy.get("bucket_name"), client=client
                )
                tasks.append(asyncio.ensure_future(task))
            for path, policy in fragment_upload_files:
                task = Client.upload_fragment_async(path, policy, client=client)
                tasks.append(asyncio.ensure_future(task))
            await asyncio.gather(*tasks)

    @staticmethod
    @execute_online_only
    def upload_fragment_sync(local_path: str, policy: dict, chunk_size: int = 1024 * 1024 * 32) -> None:
        """Upload a file in fragments."""
        upload = ResumableUpload(upload_url=policy.get("url"), chunk_size=chunk_size)
        with open(local_path, "rb") as f:
            data = f.read()
        upload._stream = io.BytesIO(data)
        upload._total_bytes = len(data)
        upload._resumable_url = policy.get("url")

        filename = os.path.basename(local_path)
        task_desc = f"Upload {filename if len(filename) <= 15 else filename[:12] + '...'}"
        task_id = Client.progress.add_task(task_desc, total=upload.total_bytes)

        bytes_uploaded = 0
        while upload.finished == False:
            method, url, payload, headers = upload._prepare_request()
            if headers.get("content-type") == None:
                headers["content-type"] = "application/octet-stream"
            result = httpx.request(method, url, data=payload, headers=headers, timeout=Client._timeout)
            upload._process_resumable_response(result, len(payload))
            Client.progress.update(task_id, advance=upload.bytes_uploaded - bytes_uploaded)
            bytes_uploaded = upload.bytes_uploaded
        Client.progress.stop_task(task_id)

    @staticmethod
    @execute_online_only
    async def upload_fragment_async(
        local_path: str,
        policy: dict,
        chunk_size: int = 1024 * 1024 * 32,
        client: Optional[httpx.AsyncClient] = None,
    ) -> None:
        """Upload a file in fragments."""
        upload = ResumableUpload(upload_url=policy.get("url"), chunk_size=chunk_size)
        with open(local_path, "rb") as f:
            data = f.read()
        upload._stream = io.BytesIO(data)
        upload._total_bytes = len(data)
        upload._resumable_url = policy.get("url")

        filename = os.path.basename(local_path)
        task_desc = f"Upload {filename if len(filename) <= 15 else filename[:12] + '...'}"
        task_id = Client.progress.add_task(task_desc, total=upload.total_bytes)

        bytes_uploaded = 0
        while upload.finished == False:
            method, url, payload, headers = upload._prepare_request()
            if headers.get("content-type") == None:
                headers["content-type"] = "application/octet-stream"
            result = await (client or httpx).request(
                method, url, data=payload, headers=headers, timeout=Client._timeout
            )
            upload._process_resumable_response(result, len(payload))
            Client.progress.update(task_id, advance=upload.bytes_uploaded - bytes_uploaded)
            bytes_uploaded = upload.bytes_uploaded
        Client.progress.stop_task(task_id)

    @staticmethod
    @execute_online_only
    def upload_single_sync(local_path: str, policy: dict, bucket: str) -> ResponseType:
        """Upload a file in a single request synchronously."""
        url = policy.get("url")
        num_bytes = os.path.getsize(local_path)
        with open(local_path, "rb") as content:
            files = {"file": (bucket, content)}
            upload_method = request_handle(httpx.post)

            filename = os.path.basename(local_path)
            task_desc = f"Upload {filename if len(filename) <= 15 else filename[:12] + '...'}"
            task_id = Client.progress.add_task(task_desc, total=num_bytes)

            res = upload_method(url, data=policy.get("fields"), files=files)
        Client.progress.update(task_id, advance=num_bytes)
        Client.progress.stop_task(task_id)
        return res

    @staticmethod
    @execute_online_only
    async def upload_single_async(
        local_path: str,
        policy: dict,
        bucket: str,
        client: Optional[httpx.AsyncClient] = None,
    ) -> ResponseType:
        """Upload a file in a single request asynchronously."""
        url = policy.get("url")
        with open(local_path, "rb") as content:
            num_bytes = os.path.getsize(local_path)
            files = {"file": (bucket, content)}
            upload_method = client.post if client else request_handle(httpx.post)
            filename = os.path.basename(local_path)
            task_desc = f"Upload {filename if len(filename) <= 15 else filename[:12] + '...'}"
            task_id = Client.progress.add_task(task_desc, total=num_bytes)
            res = await upload_method(url, data=policy.get("fields"), files=files)
        Client.progress.update(task_id, advance=num_bytes)
        Client.progress.stop_task(task_id)
        return res

    @staticmethod
    def download_sync(url: str, local_path: str, num_bytes: Optional[int] = None) -> str:
        """Download a file synchronously.

        Args:
            url (str): URL of the file to download.
            local_path (str): Local path to save the file (includes the file name)
            num_bytes (Optional[int], optional): Number of bytes to download. Default: None.
        """

        os.makedirs(os.path.dirname(local_path), exist_ok=True)
        with open(local_path, "wb") as f:
            filename = os.path.basename(local_path)
            task_desc = f"Download {filename if len(filename) <= 15 else filename[:12] + '...'}"
            task_id = Client.progress.add_task(task_desc, total=num_bytes)
            with httpx.stream("GET", url) as r:
                for data in r.iter_bytes():
                    f.write(data)
                    Client.progress.update(task_id, advance=len(data))
        return local_path

    @staticmethod
    async def download_async(
        url: str,
        local_path: str,
        num_bytes: Optional[int] = None,
        client: Optional[httpx.AsyncClient] = None,
    ) -> str:
        """Download a file asynchronously.

        Args:
            url (str): URL of the file to download.
            local_path (str): Local path to save the file (includes the file name)
            num_bytes (Optional[int], optional): Number of bytes to download. Default: None.
            client: (httpx.Client, optional): Alternative client to use (may be AsyncClient). Default: None.
            progress: (rich.progress.Progress, optional): Progress object for track download. Default: None.
        """
        os.makedirs(os.path.dirname(local_path), exist_ok=True)
        with open(local_path, "wb") as f:
            filename = os.path.basename(local_path)
            task_desc = f"Download {filename if len(filename) <= 15 else filename[:12] + '...'}"
            task_id = Client.progress.add_task(task_desc, total=num_bytes)
            async with (client or httpx.AsyncClient()).stream("GET", url) as r:
                async for data in r.aiter_bytes():
                    f.write(data)
                    Client.progress.update(task_id, advance=len(data))
        Client.progress.stop_task(task_id)
        return local_path

    @staticmethod
    def download_many_sync(urls: List[str], local_paths: List[str], num_bytes: List[int], **kwargs) -> None:
        """Download many files synchronously.

        Args:
            urls (List[str]): URLs of the files to download.
            local_paths (List[str]): Local paths to save the files (includes the file names)
            num_bytes (List[int]): Number of bytes to download.
        """
        for url, local_path, byt in zip(urls, local_paths, num_bytes):
            Client.download_sync(url, local_path, byt)

    @staticmethod
    async def download_many_async(urls: List[str], local_paths: List[str], num_bytes: List[int], **kwargs) -> None:
        """Download many files asynchronously.

        Args:
            urls (List[str]): URLs of the files to download.
            local_paths (List[str]): Local paths to save the files (includes the file names)
            num_bytes (List[int]): Number of bytes to download.
        """
        async with httpx.AsyncClient() as client:
            tasks = []
            for url, local_path, byt in zip(urls, local_paths, num_bytes):
                task = Client.download_async(url, local_path, byt, client=client)
                tasks.append(asyncio.ensure_future(task))
            await asyncio.gather(*tasks)

    @staticmethod
    def request(call: str, url: str, wait_response: bool = False, **kwargs) -> ResponseType:
        """Execute a request to the Hectiq Lab API."""
        url = Client.format_url(url)
        method = request_handle(partial(getattr(Client.client, call), url=url))
        return Client.execute(method=method, wait_response=wait_response, **kwargs)

    @staticmethod
    def execute(
        method: callable,
        wait_response: bool = False,
        is_async_method: bool = False,
        with_progress: bool = False,
        **kwargs,
    ) -> ResponseType:

        def execution_handler(**kwargs):
            """Execute a request in the background or in the main thread."""
            if wait_response:
                if is_async_method:
                    return asyncio.run(method(**kwargs))
                return method(**kwargs)

            def threading_method(**kwargs):
                if is_async_method:
                    return asyncio.run(method(**kwargs))
                return method(**kwargs)

            t = threading.Thread(target=threading_method, kwargs=kwargs)
            t.start()
            return

        if not with_progress:
            return execution_handler(**kwargs)

        with Client.progress_context():
            result = execution_handler(**kwargs)
        return result

    @staticmethod
    @contextmanager
    def progress_context():
        Client.initiate_progress()
        yield
        Client.reset_progress()

    @staticmethod
    def initiate_progress():
        Client._progress = ProgressHandler()
        try:
            Client._progress.start()
        except:
            msg = (
                "Could not start progress handler. "
                "This is likely caused by the rich libary, you may have many Progress instances "
                "in the same process. Try registering the root Progress instance before executing "
                "the script using `hl.register_progress(progress)`. Continuing without progress."
            )
            if msg not in Client._logged_error_messages:
                logger.warning(msg)
                Client._logged_error_messages.add(msg)

            Client._progress.progress = None

    @staticmethod
    def reset_progress():
        if Client.progress is None:
            return
        Client.progress.stop()
        Client._progress = None

    @staticmethod
    def format_url(url: str) -> str:
        if not url.startswith("http") and url[0] == "/":
            url = API_URL + url
        return url



import pytest
import os


@pytest.fixture
def path():
    return os.path.abspath(os.path.join(__file__, "../dummy/artifact.txt"))


def test_add_artifact(path):
    from pyhectiqlab import functional as hl, Run

    run = Run(rank=1, project="hectiq-ai/test")
    res = run.add_artifact(path)



import difflib
import typing

import click

"""Source: https://github.com/click-contrib/click-didyoumean/tree/master?tab=MIT-1-ov-file

Under MIT License:

Copyright (c) 2016 Timo Furrer

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

class DidYouMean:
    """
    Mixin class for click MultiCommand inherited classes
    to provide git-like *did-you-mean* functionality when
    a certain command is not registered.
    """

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:  # noqa: ANN401
        self.max_suggestions = kwargs.pop("max_suggestions", 4)
        self.cutoff = kwargs.pop("cutoff", 0.5)
        super().__init__(*args, **kwargs)  # type: ignore[call-arg]

    def resolve_command(
        self, ctx: click.Context, args: typing.List[str]
    ) -> typing.Tuple[
        typing.Optional[str], typing.Optional[click.Command], typing.List[str]
    ]:
        """
        Overrides clicks ``resolve_command`` method
        and appends *Did you mean ...* suggestions
        to the raised exception message.
        """
        try:
            return super().resolve_command(ctx, args)  # type: ignore[misc]
        except click.exceptions.UsageError as error:
            error_msg = "‼️  " + str(error)
            original_cmd_name = click.utils.make_str(args[0])
            matches = difflib.get_close_matches(
                original_cmd_name,
                self.list_commands(ctx),  # type: ignore[attr-defined]
                self.max_suggestions,
                self.cutoff,
            )
            if matches:
                fmt_matches = "\n    ".join(matches)
                error_msg += "\n\n"
                error_msg += f"🔍 Did you mean one of these?\n    {fmt_matches}"

            # raise "No such command '%s'." % original_cmd_name
            raise click.exceptions.UsageError(error_msg, error.ctx) from error

class DYMCommandCollection(DidYouMean, click.CommandCollection):
    """
    click CommandCollection to provide git-like
    *did-you-mean* functionality when a certain
    command is not found in the group.
    """


import clisync

from typing import Optional, List

from pyhectiqlab.client import Client
from pyhectiqlab.decorators import functional_alias


class Tag:

    @staticmethod
    @functional_alias("create_tag")
    @clisync.include()
    def create(
        name: str,
        project: str,
        color: Optional[str] = None,
        wait_response: bool = False,
    ):
        """Create a new tag.

        Args:
            name (str): Name of the tag.
            project (str): Project name.
            color (str, optional): Color of the tag. Default: None.
            wait_response (bool): Set to true to create sync. If False, the creation is made in background. Default: False.
        """
        json = {"name": name, "color": color, "project": project}
        return Client.post("/app/tags", wait_response=wait_response, json=json)

    @staticmethod
    @clisync.include()
    def attach_to_run(
        tags: List[str],
        run_id: str,
        project: str,
        wait_response: bool = False,
    ):
        """Attach tags to a run.

        Args:
            tags (List[str]): List of tags to attach.
            run_id (str): Run ID.
            project (str): Project name. Default: None.
            wait_response (bool): Set to true to attach sync. If False, the attachment is made in background. Default: False.
        """
        json = {"run": run_id, "tags": tags, "project": project}
        return Client.post("/app/tags/attach-to-run", wait_response=wait_response, json=json)

    @staticmethod
    @clisync.include()
    def detach_from_run(
        tag: str,
        run_id: str,
        project: str,
        wait_response: bool = False,
    ):
        """Detach a tag from a run.

        Args:
            tag (str): Tag name.
            run_id (str): Run ID.
            project (str, optional): Project name. Default: None.
            wait_response (bool): Set to true to detach sync. If False, the detachment is made in background. Default: False.
        """
        return Client.post(
            f"/app/tags/detach-from-run",
            wait_response=wait_response,
            json={"run": run_id, "tag": tag, "project": project},
        )

    @staticmethod
    @clisync.include()
    def attach_to_model(
        tags: List[str],
        model_id: str,
        project: str,
        wait_response: bool = False,
    ):
        """Attach a tag to a model.

        Args:
            tag (str): Tag name.
            model_id (str): Model ID.
            project (str, optional): Project name. Default: None.
            wait_response (bool): Set to true to attach sync. If False, the attachment is made in background. Default: False.
        """
        body = {"model": model_id, "tags": tags, "project": project}
        return Client.post("/app/tags/attach-to-model", json=body, wait_response=wait_response)

    @staticmethod
    @functional_alias("detach_tag_from_model")
    @clisync.include()
    def detach_from_model(
        tag: str,
        model_id: str,
        project: str,
        wait_response: bool = False,
    ):
        """Detach a tag from a model.

        Args:
            tag (str): Tag name.
            model_id (str): Model ID.
            project (str): Project name. Default: None.
            wait_response (bool): Set to true to detach sync. If False, the detachment is made in background. Default: False.
        """
        return Client.post(
            f"/app/tags/detach-from-model",
            wait_response=wait_response,
            json={"model": model_id, "tag": tag, "project": project},
        )

    @staticmethod
    @clisync.include()
    def attach_to_dataset(
        tags: List[str],
        dataset_id: str,
        project: str,
        wait_response: bool = False,
    ):
        """Attach a tag to a dataset.

        Args:
            tags (list[str]): List of tag names.
            dataset_id (str): Dataset ID.
            project (str): Project name.
            wait_response (bool): Set to true to attach sync. If False, the attachment is made in background. Default: False.
        """
        return Client.post(
            f"/app/tags/attach-to-dataset",
            wait_response=wait_response,
            json={"dataset": dataset_id, "tags": tags, "project": project},
        )

    @staticmethod
    @functional_alias("detach_tag_from_dataset")
    @clisync.include()
    def detach_from_dataset(
        tag: str,
        dataset_id: str,
        project: str,
        wait_response: bool = False,
    ):
        """Detach a tag from a dataset.

        Args:
            dataset_id (str): Dataset ID.
            tag (str): Tag name.
            project (str): Project name.
            wait_response (bool): Set to true to detach sync. If False, the detachment is made in background. Default: False.
        """
        return Client.post(
            f"/app/tags/detach-from-dataset",
            wait_response=wait_response,
            json={"dataset": dataset_id, "tag": tag, "project": project},
        )

    @staticmethod
    @functional_alias("list_tags")
    @clisync.include()
    def list(
        project: Optional[str] = None,
        page: Optional[int] = 1,
        limit: Optional[int] = 100,
        order_by: Optional[str] = None,
        order_direction: Optional[str] = None,
    ):
        """Get tags.

        Args:
            project (str, optional): Project name. Default: None.
            page (int, optional): Page number. Default: 1.
            limit (int, optional): Number of tags per page. Default: 100.
            order_by (str, optional): Order by field. Default: None.
            order_direction (str, optional): Order direction. Default: None.
        """
        if not project:
            return
        return Client.get(
            "/app/tags",
            wait_response=True,
            params={
                "project": project,
                "page": page,
                "limit": limit,
                "order_by": order_by,
                "order_direction": order_direction,
            },
        )



import sys
import click
from .auth import auth_group
from clisync import CliSync
from .diff import DYMCommandCollection


def main():
    import pyhectiqlab
    from pyhectiqlab import const
    from pyhectiqlab import Run, Model, Dataset
    from pyhectiqlab.project import Project
    from pyhectiqlab.artifact import Artifact
    from pyhectiqlab.tag import Tag
    from pyhectiqlab.message import Message
    from pyhectiqlab.versions import PackageVersion

    # enable block and project creation for the cli
    const.DISABLE_PROJECT_CREATION = False

    group = CliSync(module=pyhectiqlab, classes=[Run, Model, Dataset, Artifact, Tag, Project, PackageVersion, Message])
    cli = DYMCommandCollection(
        sources=[auth_group, group], help="👋 Hectiq Lab CLI. Documentation at https://docs.hectiq.ai."
    )
    # Standalone mode is False so that the errors can be caught by the runs
    try:
        cli(standalone_mode=False)
    except click.exceptions.UsageError as error:
        error_msg = str(error)
        print(error_msg)
    sys.exit()


if __name__ == "__main__":
    main()




import os
import clisync
import shutil

from typing import Optional, List, Union

from pyhectiqlab.tag import Tag
from pyhectiqlab.project import Project
from pyhectiqlab.client import Client
from pyhectiqlab.utils import is_running_event_loop
from pyhectiqlab.decorators import functional_alias
from pyhectiqlab.settings import getenv
from pyhectiqlab.logging import hectiqlab_logger as logger
from .utils import batched, list_all_files_in_dir, extract_host_from_source


class Dataset:
    _client: Client = Client

    @staticmethod
    @functional_alias("create_dataset")
    @clisync.include()
    def create(
        name: str,
        source: str,
        host: Optional[str] = None,
        description: Optional[str] = None,
        version: Optional[str] = None,
        run_id: Optional[str] = None,
        project: Optional[str] = None,
        upload: Optional[bool] = True,
    ):
        """Create a dataset.

        Supported cloud storages:
        - Amazon S3
        - Google Cloud Storage

        Args:
            name (str): Name of the dataset.
            source (str): Path to the dataset. If your dataset is located in a local directory, the source should be the directory path (e.g., "/path/to/dataset"). If
                the dataset is located in a cloud storage, the source should be the URL of the dataset (e.g., "s3://bucket/dataset").
            host (str, optional): Host of the dataset. Default: None. If the dataset is located in a local directory,
                the host could be your hostname or leave it empty. If the dataset is located in a cloud storage, the host should be the cloud storage name ("s3" or "gs")
            description (str, optional): Description of the dataset. Default: None.
            version (str]): Version of the dataset. If None, the created dataset version increments the minor of the latter version if it exists, and uses 1.0 otherwise. Default: None.
            run_id (str, optional): ID of the run to attach to the dataset. If None, the dataset is not attached. Default: None.
            project (str, optional): Project of the dataset. Default: None.
            upload (bool): If True, uploads the dataset to the Lab. Default: True.

        Raises:
            logger.error: If the `name` or `source` parameters are not provided, the project is not found or the dataset creation failed, it does nothing and returns None.

        Skips dataset creation if the project is not found, or if the dataset creation failed.
        """

        if name is None:
            logger.error("The `name` parameter is required for dataset creation.")
            return

        if source is None:
            logger.error("The `source` parameter is required for dataset creation.")
            return

        from pyhectiqlab.run import Run

        project = Project.get(project)
        run_id = Run.get_id(run_id)
        if project is None:
            msg = f"Could not create dataset {name}, project not found."
            if not Dataset._client.online():
                msg = msg[:-1] + " (offline mode)."
            msg += " Skipping dataset creation."
            logger.error(msg)
            return
        source = os.path.abspath(source)
        host = extract_host_from_source(source)

        data = {
            "name": name,
            "description": description,
            "version": version,
            "host": host,
            "source": source,
            "project": project,
            "root_run": run_id,
        }
        dataset = Dataset._client.post("/app/datasets", wait_response=True, json=data)

        if dataset is None or "id" not in dataset:
            logger.error(f"Failed to create dataset {name}.")
            return
        if upload and host not in ["s3", "gs"]:
            Dataset.upload(id=dataset["id"], source=source)

        if run_id is not None:
            Dataset.attach(name=name, version=version, run_id=run_id, project=project)
        return dataset

    @staticmethod
    @functional_alias("retrieve_dataset")
    @clisync.include()
    def retrieve(name: str, version: str, project: str, fields: Optional[List[str]] = None):
        """Retrieve a dataset

        Args:
            name (str): Name of the dataset
            project (str): Project of the dataset
            version (str): Version of the dataset
            fields (list[str], optional): Fields to retrieve. Default: None.

        Raises:
            logger.error: If the project is not found, it does nothing and returns None.
        """
        project = Project.get(project)
        if project is None:
            msg = f"Could not retrieve the dataset {name}-{version}, project not found."
            if not Dataset._client.online():
                msg = msg[:-1] + " (offline mode)."
            msg += " Skipping dataset retrieval."
            logger.error(msg)
            return
        body = Dataset._client.get(
            "/app/datasets/retrieve",
            params={
                "name": name,
                "project": project,
                "version": version,
                "fields": fields or [],
            },
            wait_response=True,
        )
        return body

    @staticmethod
    @functional_alias("dataset_exists")
    @clisync.include()
    def exists(name: str, version: str, project: Optional[str] = None):
        """Check if a dataset exists

        Args:
            name (str): Name of the dataset
            project (str): Project of the dataset
            version (str): Version of the dataset
        """
        return Dataset.retrieve(name=name, project=project, version=version, fields=["id"]) is not None

    @staticmethod
    @functional_alias("list_datasets")
    @clisync.include()
    def list(
        project: Optional[str] = None,
        search: Optional[str] = None,
        author: Optional[str] = None,
        keep_latest_version: bool = False,
        fields: Optional[List[str]] = [],
        page: Optional[int] = 1,
        limit: Optional[int] = 100,
        order_by: Optional[str] = "created_at",
        order_direction: Optional[str] = "desc",
    ):
        """List the datasets

        Args:
            project (str, optional): Project of the dataset.
            search (str, optional): Search string.
            author (str, optional): Author of the dataset.
            keep_latest_version (bool, optional): If True, group by the latest version of dataset name and return only the latest version of each dataset name.
            fields (list[str], optional): Fields to retrieve.
            page (int, optional): Page number.
            limit (int, optional): Limit of the datasets.
            order_by (str, optional): Order by.
            order_direction (str, optional): Order direction.
            wait_response (bool): Wait for the response from the server. Default: False.
        """
        params = {
            "project": project,
            "search": search,
            "author": author,
            "keep_latest_version": keep_latest_version,
            "fields": fields,
            "page": page or 1,
            "limit": limit or 100,
            "order_by": order_by,
            "order_direction": order_direction,
        }

        return Dataset._client.get("/app/datasets", params=params, wait_response=True)

    @staticmethod
    @functional_alias("update_dataset")
    @clisync.include(wait_response=True)
    def update(
        id: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        version: Optional[str] = None,
        wait_response: bool = False,
    ):
        """Update a dataset, with new name, description, and/or version.

        Args:
            id (str): ID of the dataset.
            name (str, optional): Name of the dataset.
            description (str, optional): Description of the dataset.
            version (str, optional): Version of the dataset.
            wait_response (bool): Wait for the response from the server. Default: False

        Raises:
            logging.error: If the `id` parameter is not provided, it does nothing and returns None.
        """
        if not id:
            logger.error("The `id` parameter is required.")
            return
        return Dataset._client.put(
            f"/app/datasets/{id}",
            json={"name": name, "description": description, "version": version},
            wait_response=wait_response,
        )

    @staticmethod
    @functional_alias("delete_dataset")
    @clisync.include(wait_response=True)
    def delete(
        id: Optional[str] = None,
        name: Optional[str] = None,
        version: Optional[str] = None,
        project: Optional[str] = None,
        wait_response: bool = False,
    ):
        """Delete a dataset.

        Args:
            id (str): ID of the dataset.
            name (str, optional): Name of the dataset.
            version (str, optional): Version of the dataset.
            project (str, optional): Project of the dataset.
            wait_response (bool): Wait for the response from the server. Default: False

        Raises:
            logging.error: If the `id` is not provided, or if all `name`, `version`, `project` parameters are not provided, or if the dataset is not found, it does nothing and returns None.
        """
        if id is None or (name is None or version is None or project is None):
            logger.error("The `id` or `name`, `version`, `project` parameters are required.")
            return

        if not id:
            dataset = Dataset.retrieve(name=name, project=project, version=version, fields=["id"])
            if not dataset:
                logger.error(f"Could not delete dataset {name}-{version}, dataset not found. Skipping deletion.")
                return
            id = dataset["id"]
        return Dataset._client.delete(f"/app/datasets/{id}", wait_response=wait_response)

    @staticmethod
    @clisync.include()
    @functional_alias("upload_dataset")
    @clisync.include()
    def upload(id: str, source: str) -> None:
        """Upload local files to a dataset hosted in the Lab.

        Args:
            id (str): ID of the dataset.
            path (str): Path to the directory containing the files to upload.
        """
        all_files = list_all_files_in_dir(source) if os.path.isdir(source) else [source]
        batch_size = 50
        Dataset._id = dataset_id = id
        def composition_sync(**kwargs):
            for batch in batched(all_files, batch_size):
                # Create many files
                batch_body = [
                    {"name": os.path.relpath(f, start=source if os.path.isdir(source) else os.path.dirname(f)), 
                     "num_bytes": os.path.getsize(f)} for f in batch
                ]
                files = Dataset._client.post(
                    f"/app/datasets/{dataset_id}/files", wait_response=True, json={"files": batch_body}
                )["results"]

                # For each file in batch, get the policy in files (use name for finding the file)
                sorted_files = []
                for el in batch:
                    file = [
                        f.get("upload_policy")
                        for f in files
                        if f["name"]
                        == os.path.relpath(el, start=source if os.path.isdir(source) else os.path.dirname(el))
                    ][0]
                    sorted_files.append(file)
                # Upload dataset files
                Dataset._client.upload_many_sync(paths=batch, policies=sorted_files, **kwargs)

        async def composition_async(**kwargs):
            for batch in batched(all_files, batch_size):
                # Create many files
                batch_body = [
                    {"name": os.path.relpath(f, start=source if os.path.isdir(source) else os.path.dirname(f)), 
                     "num_bytes": os.path.getsize(f)} for f in batch
                ]
                files = Dataset._client.post(
                    f"/app/datasets/{dataset_id}/files", wait_response=True, json={"files": batch_body}
                )["results"]

                # For each file in batch, get the policy in files (use name for finding the file)
                sorted_files = []
                for el in batch:
                    name = os.path.relpath(el, start=source if os.path.isdir(source) else os.path.dirname(el))
                    file = [
                        f.get("upload_policy")
                        for f in files
                        if f["name"]
                        == name
                    ][0]
                    sorted_files.append(file)
                # Upload dataset files
                await Dataset._client.upload_many_async(paths=batch, policies=sorted_files, **kwargs)

        if is_running_event_loop():
            Dataset._client.execute(composition_sync, wait_response=True, is_async_method=False, with_progress=True)
        else:
            Dataset._client.execute(composition_async, wait_response=True, is_async_method=True, with_progress=True)

    @staticmethod
    @functional_alias("download_dataset")
    @clisync.include()
    def download(
        name: str,
        version: str,
        project: Optional[str] = None,
        path: Optional[str] = None,
        overwrite: bool = False,
    ):
        """Download a dataset synchronously

        Args:
            name (str): Name of the dataset.
            version (str): Version of the dataset.
            project (str, optional): Project of the dataset.
            path (str): Path to download the dataset. If None, it uses the `HECTIQLAB_DATASETS_DOWNLOAD` environment variable or the current directory.
            overwrite (bool): Overwrite the existing files. Default: False.

        Raises:
            logger.error: If the project is not found, or if the dataset is not found, it does nothing and returns None.
        """

        from pyhectiqlab import Run

        project = Project.get(project)
        if project is None:
            msg = f"Could not download dataset {name}-{version}, project not found."
            if not Dataset._client.online():
                msg = msg[:-1] + " (offline mode)."
            msg += " Skipping dataset download."
            logger.error(msg)
            return
        if not Dataset.exists(name=name, version=version, project=project):
            logger.error(
                f"Could not download dataset {name}-{version}, dataset not found. Skipping datgaset download."
            )
            return

        run_id = Run.get_id()
        if run_id is not None:
            Dataset.attach(name=name, version=version, run_id=run_id, project=project)

        path = path or getenv("HECTIQLAB_DATASETS_DOWNLOAD", ".")
        if not os.path.basename(path) == f"{name}-{version}":
            path = os.path.join(path, f"{name}-{version}")
        if os.path.exists(path):
            if not overwrite:
                logger.warning(
                    f"Directory {path} already exists. Set `overwrite=True` to overwrite the files. Skipping dataset download."
                )
                return path
            shutil.rmtree(path)

        os.makedirs(path, exist_ok=True)

        def composition_sync(**kwargs):
            dataset = Dataset.retrieve(name=name, project=project, version=version, fields=["id"])
            if not dataset:
                return
            dataset_id = dataset["id"]

            num_results = 0
            total_results = 1
            page = 1
            while not num_results == total_results:
                files = Dataset._client.get(
                    f"/app/datasets/{dataset_id}/files/",
                    params={"page": page, "limit": 50, "fields": ["download_url", "name", "num_bytes"], "flat": True},
                    wait_response=True,
                )
                total_results = files["total_results"]

                Dataset._client.download_many_sync(
                    urls=[file["download_url"] for file in files["results"]],
                    local_paths=[os.path.join(path, file["name"]) for file in files["results"]],
                    num_bytes=[file["num_bytes"] for file in files["results"]],
                    **kwargs,
                )
                num_results += len(files["results"])
                page += 1

            return Dataset._client.post(f"/app/datasets/{dataset_id}/on-download-completed", wait_response=False)

        async def composition_async(**kwargs):
            dataset = Dataset.retrieve(name=name, project=project, version=version, fields=["id"])
            if not dataset:
                return
            dataset_id = dataset["id"]

            num_results = 0
            total_results = 1
            page = 1
            while not num_results == total_results:
                files = Dataset._client.get(
                    f"/app/datasets/{dataset_id}/files/",
                    params={"page": page, "limit": 50, "fields": ["download_url", "name", "num_bytes"], "flat": True},
                    wait_response=True,
                )
                total_results = files["total_results"]

                await Dataset._client.download_many_async(
                    urls=[file["download_url"] for file in files["results"]],
                    local_paths=[os.path.join(path, file["name"]) for file in files["results"]],
                    num_bytes=[file["num_bytes"] for file in files["results"]],
                    **kwargs,
                )
                num_results += len(files["results"])
                page += 1

            return Dataset._client.post(f"/app/datasets/{dataset_id}/on-download-completed", wait_response=False)

        logger.info(f"⏳ Downloading dataset {name}-{version} to {path}...")
        if is_running_event_loop():
            Dataset._client.execute(composition_sync, wait_response=True, is_async_method=False, with_progress=True)
        else:
            Dataset._client.execute(composition_async, wait_response=True, is_async_method=True, with_progress=True)
        logger.info("✅ Done.")
        return path

    @staticmethod
    @clisync.include()
    def attach(
        name: str,
        version: str,
        run_id: str,
        project: Optional[str] = None,
        wait_response: bool = False,
    ):
        """Attach a dataset to a run

        Args:
            name (str): Name of the dataset.
            version (str): Version of the dataset.
            run_id (str): ID of the run.
            project (str, optional): Project of the dataset.
            wait_response (bool): Wait for the response from the server. Default: False.

        Raises:
            logger.error:  If the `run_id` parameter is not provided or the dataset is not found, it does nothing and returns None.
        """
        if run_id is None:
            logger.error("The `run_id` parameter is required.")
            return
        dataset = Dataset.retrieve(name=name, project=project, version=version, fields=["id"])
        if dataset is None:
            logger.error(
                f"Could not attach run {run_id}, dataset {name}-{version} not found. Skipping attachment to run."
            )
            return
        return Dataset._client.post(
            f"/app/datasets/{dataset['id']}/attach-to-run/{run_id}", wait_response=wait_response
        )
    
    @staticmethod
    @clisync.include()
    def detach(
        name: str,
        version: str,
        run_id: str,
        project: Optional[str] = None,
        wait_response: bool = False,
    ):
        """Detach a dataset from a run

        Args:
            name (str): Name of the dataset.
            version (str): Version of the dataset.
            run_id (str): ID of the run.
            project (str, optional): Project of the dataset.
            wait_response (bool, optional): Wait for the response from the server. Default: False.

        Raises:
            ValueError: If the `run_id` parameter is not provided.


        Raises:
            logger.error:  If the `run_id` parameter is not provided or the dataset is not found, it does nothing and returns None.
        """
        if run_id is None:
            logger.error("The `run_id` parameter is required.")
            return
        dataset = Dataset.retrieve(name=name, project=project, version=version, fields=["id"])
        if not dataset:
            logger.error(
                f"Could not detach run {run_id}, dataset {name}-{version} not found. Skipping detachment from run."
            )
            return
        dataset_id = dataset["id"]
        return Dataset._client.post(
            f"/app/datasets/{dataset_id}/detach-from-run/{run_id}", wait_response=wait_response
        )

    @staticmethod
    @functional_alias("add_tags_to_dataset")
    @clisync.include()
    def add_tags(
        tags: Union[str, List[str]],
        name: str,
        version: str,
        project: Optional[str] = None,
        wait_response: Optional[bool] = False,
    ):
        """Add tags to a dataset

        Args:
            tags (Union[str, List[str]]): Tags to add.
            name (str): Name of the dataset.
            version (str): Version of the dataset.
            project (str, optional): Project name. Default: None.
            wait_response (bool, optional): Wait for the response from the server. Default: False.

        Raises:
            logger.error:  If the dataset is not found, it does nothing and returns None.
        """
        project = Project.get(project)
        dataset = Dataset.retrieve(name=name, project=project, version=version, fields=["id"])
        if not dataset:
            logger.error(
                f"Could not attach tags {tags}, dataset {name}-{version} not found. Skipping attachment of tags."
            )
            return
        dataset_id = dataset["id"]
        return Tag.attach_to_dataset(tags=tags, dataset_id=dataset_id, project=project, wait_response=wait_response)

    @staticmethod
    @functional_alias("detach_tag_from_dataset")
    @clisync.include()
    def detach_tag(
        tag: str,
        name: str,
        version: str,
        project: Optional[str] = None,
        wait_response: Optional[bool] = False,
    ):
        """Remove a tag from the run.
        For functional alias, use `detach_tag_from_run`.

        Args:
            title (str): The new title of the run.
            name (str): Name of the dataset.
            version (str): Version of the dataset.
            project (str, optional): Project name. Default: None.

        Raises:
            logger.error:  If the dataset is not found, it does nothing and returns None.
        """
        project = Project.get(project)
        dataset = Dataset.retrieve(name=name, version=version, project=project, fields=["id"])
        if not dataset:
            logger.error(
                f"Could not detach tag {tag}, dataset {name}-{version} not found. Skipping detachment of tag."
            )
            return
        dataset_id = dataset["id"]
        Tag.detach_from_dataset(tag=tag, dataset_id=dataset_id, project=project, wait_response=wait_response)
